// PV112 2017, lesson 2

// This lesson has 10 tasks; marked "TASK 1" etc. Occasionally, there are more
// than one task label per task and file. We will follow and solve these tasks
// throughout the lesson, explaining any new concepts and principles. Feel free
// to work faster than the tutor, but make sure you understand the principles.
// The task descriptions offer only basic hints on their solution, so you will
// need to read documentation or lecture slides. Or, of course, you can wait
// until the tutor catches up and solves the task while providing explanation.

#include <iostream>
#include "PV112.h"

#define _USE_MATH_DEFINES
#include <cmath>

// Include GLEW, use static library
#define GLEW_STATIC
#include "GL/glew.h"
#if defined(_WIN32)
#define NOMINMAX // Make Windows.h not define 'min' and 'max' macros
#include <GL/wglew.h> // Include on Windows
#pragma comment(lib, "glew32s.lib")
#else
#include <GL/glxew.h> // Include on Linux and Mac
#define APIENTRY
#endif

#include <GL/freeglut.h>

//#include <glm/glm.hpp> // Perhaps not needed
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/matrix_transform.hpp>

using namespace std;
using namespace PV112;

// Current window size
int win_width = 480;
int win_height = 480;

GLuint program;

GLint aspect_loc;
GLint color_loc;
GLint model_position_loc;
GLint model_matrix_loc;
GLint view_matrix_loc;
GLint projection_matrix_loc;
GLint PVM_matrix_loc;
GLint normal_matrix_loc;

// Simple geometries that we will use in this lecture
PV112Geometry my_cube;
PV112Geometry my_sphere;
PV112Geometry my_teapot;

// Current time of the application in seconds, for animations
float app_time_s = 0.0f;

// Called when the user presses a key
void key_pressed(unsigned char key, int mouseX, int mouseY)
{
    switch (key)
    {
    case 27: // Escape
     exit(0);
     break;
    case 'l':
     glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
     glutPostRedisplay();
     break;
    case 'f':
     glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
     glutPostRedisplay();
     break;
    case 't':
     glutFullScreenToggle();
     break;
    }
}

// Initializes OpenGL stuff
void init()
{
    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glClearDepth(1.0);
    // TASK 1: Make wireframe mode default. You can change thickness of the
    // lines, if you want to.
    /*
    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    glLineWidth(2.0f);
    */

    // glEnable(GL_DEPTH_TEST);

    // Create shader program
    program = CreateAndLinkProgram("vertex.glsl", "fragment.glsl");
    if (0 == program)
     WaitForEnterAndExit();

    // Get attribute and uniform locations
    int position_loc = glGetAttribLocation(program, "position");
    int normal_loc = glGetAttribLocation(program, "normal"); // No matter that it is not present, it is -1 if not present
    aspect_loc = glGetUniformLocation(program, "aspect");
    color_loc = glGetUniformLocation(program, "color");
    model_position_loc = glGetUniformLocation(program, "model_position");
    model_matrix_loc = glGetUniformLocation(program, "model_matrix");
    view_matrix_loc = glGetUniformLocation(program, "view_matrix");
    projection_matrix_loc = glGetUniformLocation(program, "projection_matrix");
    PVM_matrix_loc = glGetUniformLocation(program, "PVM_matrix");
    normal_matrix_loc = glGetUniformLocation(program, "normal_matrix");

    my_cube = CreateCube(position_loc, normal_loc);
    my_sphere = CreateSphere(position_loc, normal_loc);
    my_teapot = CreateTeapot(position_loc, normal_loc);
}

// Called when the window needs to be rerendered
void render()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    // Some matrices
    // TASK 5: Use these matrices instead of the coordinate transformations
    // previously used in shaders. Useful functions to look up are:
    // glm::translate, glm::lookAt, glm::perspective, glm::radians
    // Pass them to shaders with glUniformMatrix4fv(loc, 1, GL_FALSE, glm::value_ptr(matrix))
    // This task is kind of a biggie.

    // TASK 6: Multiply them in C++ code and pass only PVM_matrix
    glm::mat4 projection_matrix, view_matrix, model_matrix, PVM_matrix;

    // TASK 7: Simple light. You don't need to understand this, it will be on
    // a lecture later. For each object, create  a normal matrix as follows:
    //normal_matrix = glm::inverse(glm::transpose(glm::mat3(view_matrix * model_matrix)));
    // Don't forget to pass it to the shaders. Note it is a 3x3 matrix.
    glm::mat3 normal_matrix;

    glUseProgram(program);

    glUniform1f(aspect_loc, float(win_width) / float(win_height));

    //------------
    //-- CUBE --

    glBindVertexArray(my_cube.VAO);
    glUniform3f(color_loc, 1.0f, 1.0f, 1.0f);
    //glUniform3f(model_position_loc, 0.0f, 0.0f, 0.0f);
    DrawGeometry(my_cube);

    // TASK 10 (optional): Using non-uniform scale on cube, create a simple
    // desk. You don't have to create new cubes, just make more draw calls with
    // different matrices. Place two spheres and a rotating teapot on top of
    // the desk. Feel free to play with the transformations; e.g. let one of
    // the spheres jump on the table and the other one change its size over
    // time. You can utilize for example the sine function.

    //--------------
    //-- SPHERE --

    // TASK 2: Add this sphere. It is supposed to be behind the cube, but it
    // doesn't seem so. You need to enable depth buffer in init() first.
    /*
    glBindVertexArray(my_sphere.VAO);
    glUniform3f(color_loc, 1.0f, 0.0f, 0.0f);
    glUniform3f(model_position_loc, 1.0f, 0.0f, -2.0f);
    DrawGeometry(my_sphere);
    */

    //--------------
    //-- TEAPOT --

    // TASK 8: Draw this teapot instead of the sphere above. Move it out of the
    // cube. Using glm::rotate and our clock variable, 'app_time_s', make the
    // teapot rotate. You can combine two transformations by multiplying their
    // matrices. What happens, when you swap the order of the matrices? Why?
    // Also, now it is a good time to disable the wireframe mode in init().

    // TASK 9: Try resizing the teapot using glm::scale. Try also non-uniform
    // scaling (i.e. every axis has a different scaling factor). In which order
    // should you multiply the matrices then? Try to understand what and why
    // happens when you reorder them.
    /*
    glBindVertexArray(my_teapot.VAO);
    glUniform3f(color_loc, 0.0f, 0.0f, 1.0f);
    DrawGeometry(my_teapot);
    */

    glBindVertexArray(0);
    glUseProgram(0);

    // Swaps the front and back buffer (double-buffering)
    glutSwapBuffers();
}

// Called when the window changes its size
void reshape(int width, int height)
{
    win_width = width;
    win_height = height;

    // Set the area into which we render
    glViewport(0, 0, win_width, win_height);
}

// Callback function to be called when we make an error in OpenGL
void APIENTRY simple_debug_callback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const char* message, const void* userParam)
{
    switch (type) // Report only some errors
    {
    case GL_DEBUG_TYPE_ERROR:
    case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR:
    case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR:
     cout << message << endl; // Put the breakpoint here
     return;
    default:
     return;
    }
}

// Simple timer function for animations
void timer(int)
{
    // Add 20 ms to the application time, register the timer to be called again in 20 ms, and rerender the scene
    app_time_s += 0.020f;
    glutTimerFunc(20, timer, 0);
    glutPostRedisplay();
}

int main(int argc, char ** argv)
{
    // Initialize GLUT
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
    glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS); // Return from the main loop

    // Set OpenGL Context parameters
    glutInitContextVersion(3, 3); // Use OpenGL 3.3
    glutInitContextProfile(GLUT_CORE_PROFILE); // Use OpenGL core profile
    glutInitContextFlags(GLUT_DEBUG); // Use OpenGL debug context

    // Initialize window
    glutInitWindowSize(win_width, win_height);
    glutCreateWindow("PV112 - cv2");

    // Initialize GLEW
    glewExperimental = GL_TRUE;
    glewInit();

    SetDebugCallback(simple_debug_callback);

    // Initialize our OpenGL stuff
    init();

    // Register callbacks
    glutDisplayFunc(render);
    glutReshapeFunc(reshape);
    glutKeyboardFunc(key_pressed);
    glutTimerFunc(20, timer, 0);

    // Run the main loop
    glutMainLoop();

    return 0;
}
